/*

    oANSIcvt.c      Converts ANIS screens to Opus oANSI

      Ŀ
                                 OANSICVT                           
                    Program to convert ANSI files to oANSI          
      

           This program is Copyright (C) 1987 David L Nugent.
             DO NOT REMOVE OR MODIFY THIS COPYRIGHT NOTICE
     If you do, may your conscience haunt you for the rest of your life.

     Version 0.1:  Nov 30 1987, Initial version with translate for all
                   standard ANSI commands implimented in the standard
		   PC/MS-DOS ANSI.SYS.

     Version 0.2:  Dec 14, 1987.  Converted to C with TP14 plus a few
		   edits for better optimisation.  Version not released.

     Version 0.3:  Jan 13, 1987.  Employed control character 'escaping'
		   technique used by Opus 1.03a to avoid problems with
		   special characters interpreted by DOS.  Added test
		   to insert CR/LF in preference to locate at beginning
		   of next line (good for non-ANSI .BBS files).  Used
		   save/restore cursor position as ignore/unignore chrs
		   as this technique is used by ANSI drawers to insert
		   artificial line breaks.

     Version 0.4:  Jan 14, 1985.  Added colour optimisation, accept
		   all colour codings and testing current pallet before
		   changing.  Significant space savings and optimisation
		   achieved on output of ANSPAINT and THEDRAW pics.

     Version 0.5:  Jan 14, 1988.  Commented out progress report to achieve
		   speed increase (by 1000s of %).  It was really intended
		   for debugging anyway.

*/

#ifndef LINT_ARGS
#define LINT_ARGS
#endif

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <ctype.h>

typedef enum {
    Up,
    Down,
    Left,
    Right
} direction;

typedef char bool;

#define TRUE    1               /* Boolean values */
#define FALSE   0

static char  ANSI_LEAD  = '\033';    /*  Escape character lead-in  */
static char  ANSI_CMD   = '[';       /*  Second ANSI lead in chr   */
static char  ANSI_DELM  = ';';       /*  ANSI Parameter delimiter  */
static char  OANS_LEAD  = '\026';    /*  oANSI video lead-in       */
static char  OANS_ATTR  = '\001';    /*  set colour attribute      */
static char  OANS_BLNK  = '\002';    /*  set blinking attribute    */
static char  OANS_UP    = '\003';    /*  cursor up one line        */
static char  OANS_DOWN  = '\004';    /*  cursor down one line      */
static char  OANS_LEFT  = '\005';    /*  cursor left one char      */
static char  OANS_RGHT  = '\006';    /*  cusror right one char     */
static char  OANS_CLRE  = '\007';    /*  clear to end of line      */
static char  OANS_POSN  = '\010';    /*  set cursor position       */
static char  OANS_CLRS  = '\014';    /*  clear screen/home cursor  */
static char  OANS_NMRE  = '\013';    /*  Turns auto-more off       */
static char  OANS_MORE  = '\005';    /*  Turns auto-more on        */
static char  OANS_ESC   = '\020';    /*  oANSI escape for ctrl chr */

#define BLACK   0                   /*  oANSI & PC colour codes  */
#define BLUE    1
#define GREEN   2
#define CYAN    3
#define RED     4
#define MAGENTA 5
#define YELLOW  6
#define WHITE   7
#define BRIGHT  8                   /*  Value to add if intensity on  */

#define DULL    0
#define BLINK   TRUE                /*  Value to add if blinking  */
#define NOBLINK FALSE

#define CTRLC   3                   /*  Special escaped characters  */
#define BACKSP  8
#define TAB     9
#define LF     10
#define CR     13
#define CTRLQ  17
#define CTRLS  19
#define CTRLZ  26
#define SPACE  32

		    /*  Program variables  */

FILE *infile, *outfile;
char infilename[64], outfilename[64];
char codebuff[60];
int  ccbuff;

int  foreground=CYAN;
int  background=BLACK;
int  oldattr=CYAN;
char blinkattr=NOBLINK;
char blinkstate=FALSE;
int  intensity=DULL;
bool active=TRUE;
bool animate=FALSE;

int  ypos=0, xpos=0;
int  ysav=0, xsav=0;

int  margc;
char **margv;


  /* ======================= Main Program Segment =========================== */


void copyright()            /* display copyright message to console */
{

    fprintf(stderr,"oANSICVT:  oANSI File Converter     Version 0.5\n");
    fprintf(stderr,"This program is Copyrighted (c) 1987, 1988 by David Nugent\n");
} 


int getinfilename()         /*  Get input file: returns no of chars  */
{

    fprintf(stdout,"File to Convert: ");
    gets(infilename);
    return strlen(infilename);
} 


int getoutfilename()        /*  Get output file: returns no of chars  */
{

    fprintf(stdout," Output to file: ");
    gets(outfilename);
    return strlen(outfilename);
} 


char  getfilenames()        /*  Retrieve file names, from command line
				or ask user, if no parameters provided,
				open each or return FALSE if error or
				no input provided by user, ie. empty   */
{ 
    bool  ingore, retval;

    retval = FALSE;         /*  Default return value (nothing entered) */
    ingore = FALSE;
    if ((margc-1) > 0) strcpy(infilename, margv[1]);
		  else ingore = (!getinfilename());
    if (!ingore) {
	if ((infile = fopen(infilename,"rb")) == NULL) {
	    fprintf(stderr,"Can't open input file: ",infilename);
	    exit(1);
	}
	else {
	    if ((margc-1) > 1) strcpy(outfilename, margv[2]);
			  else ingore = (!getoutfilename());
	    if (!ingore) {
		if ((outfile = fopen(outfilename,"wb")) == NULL) {
		    fclose(infile);
		    fprintf(stderr,"Can't create output file '%s'\n",outfilename);
		    exit(2);
		}
		else retval = TRUE;
	    }
	}
    }
    return retval;
} 


char getchr()
{

    return fgetc(infile);
} 


void checkpos()                         /*  Check and adjust positions  */
{

    while (xpos > 80) {
	ypos++;
	xpos -= 80;
    }
    while (xpos < 0) {
	ypos--;
	xpos += 80;
    } 
    if (ypos < 0) ypos = 0;
    if (ypos > 24) ypos = 24;
} 


void putchr(ch, move)
unsigned char ch;
bool          move;                     /*  Put next character  */
{
    unsigned char attr;
    void attrset(unsigned char);

    if (!active) return;
    if (move) {
	attr = foreground + intensity + (background*16);
	if (ch == SPACE) {       /* Test only background if space character */
	    if (background != oldattr/16 || blinkstate && !blinkattr)
		attrset(attr);
	}
	else {                  /* Otherwise, test the complete attribute */
	    if (attr != oldattr || (blinkstate && !blinkattr))
		attrset(attr);
	}
	if (blinkattr && !blinkstate) {
	    putchr(OANS_LEAD, FALSE);
	    putchr(OANS_BLNK, FALSE);
	    blinkstate = TRUE;
	}
    }
    fputc(ch,outfile);
    if (move) {
	switch (ch) {                   /*  MOVEMENT PROCESSING  */
	  case BACKSP:  xpos--;         /*  Backspace  */
			break;
	      case CR:  xpos = 0;       /*  Carriage return  */
			break;
	      case LF:  ypos++;         /*  Line feed  */
			break;
	     case TAB:  do {            /*  Horizontal tab 8  */
			    xpos++;
			}  while (xpos % 8);
			break;
	      default:  xpos++;
        break; 
        } 
	checkpos();
    }
} 


void attrset(attr)
unsigned char attr;
{

    putchr(OANS_LEAD, FALSE);
    putchr(OANS_ATTR, FALSE);
    if (attr < 32) {
	putchr(OANS_ESC, FALSE);         /* Escape ctrl characters */
	putchr(attr | 0x80, FALSE);
    }
    else putchr(attr, FALSE);
    blinkstate = FALSE;
    oldattr = attr;
}



void putattr(a)
int a;                          /*  Put characters for attribute change  */
{ 

    switch (a) {                /*   SET GENERAL ATTRIBUTES  */
	case 0: intensity   = DULL;         /*   0 resets all modes  */
		blinkattr   = NOBLINK;
		foreground  = WHITE;
		background  = BLACK;
		break;
	case 1: intensity   = BRIGHT;       /*   1 makes foreground bright  */
		break;
	case 4: foreground  = BLUE;         /*   4 mono underline  */
		break;
	case 5: blinkattr   = BLINK;        /*   5 makes characters blink  */
		break;
	case 7: foreground  = BLACK;        /*   7 makes characters reverse  */
		background  = WHITE;
		intensity   = DULL;
		break;
	case 8: foreground  = BLACK;        /*   8 suppressed display  */
		background  = BLACK;
		intensity   = DULL;
		break;
                                               /*  SET COLOUR ATTRIBUTES  */ 
       case 30: foreground  = BLACK;
		break;
       case 31: foreground  = RED;
		break;
       case 32: foreground  = GREEN;
		break;
       case 33: foreground  = YELLOW;
		break;
       case 34: foreground  = BLUE;
		break;
       case 35: foreground  = MAGENTA;
		break;
       case 36: foreground  = CYAN;
		break;
       case 37: foreground  = WHITE;
		break;
       case 40: background  = BLACK;
		break;
       case 41: background = RED;
		break;
       case 42: background = GREEN;
		break;
       case 43: background = YELLOW;
		break;
       case 44: background = BLUE;
		break;
       case 45: background = MAGENTA;
		break;
       case 46: background = CYAN;
		break;
       case 47: background = WHITE;
		break;
    }
}


void setattr()                      /*  Output oANSI set attributes  */
{
    int  n;
    unsigned char c, *s=codebuff;

    while (*s) {
	n = atoi(s);
	putattr(n);
	do s++; while (isdigit(*s));
	s++;
    }
} 


void putposn()
{

    putchr(OANS_LEAD, FALSE);
    putchr(OANS_POSN, FALSE);
    putchr(ypos, FALSE);
    putchr(xpos, FALSE);
    checkpos();
} 


void setposn()                          /*  Output oANSI set cursor posn  */
{ 
    int     posy, posx;
    char    *s=codebuff;

    posy = atoi(s);
    while (isdigit(*s)) s++;
    posx = atoi(s);
    if (posy) posy--;
    if (posx) posx--;
    if ((posx < 1) && (posy == ypos+1)) {
	putchr('\r');
	putchr('\n');
    }
    else if ((ypos != posy) || (xpos != posx)) {
	ypos = posy;
	xpos = posx;
	checkpos();
	putposn();
    }
} 


void clrscreen()                            /*  Output oANSI clear screen  */
{

    putchr(OANS_CLRS, FALSE);
    ypos = 0;
    xpos = 0;
} 


void clrtoend()                     /*  Output oANSI clear to end of line  */
{

    putchr(OANS_LEAD, FALSE);
    putchr(OANS_CLRE, FALSE);
} 


void cursormove(d)
direction    d;                         /*  Output oANSI cursor movement  */
{
    register int i;
    int  n, x=0;

    n = atoi(codebuff);
    if (n < 1) n = 1;
    for (i = 1; i <= n; i++) {
	switch (d) {
	  case Up: putchr(OANS_LEAD, FALSE);
		   putchr(OANS_UP, FALSE);
		   ypos--;
		   break;
	case Down: putchr(OANS_LEAD, FALSE);
		   putchr(OANS_DOWN, FALSE);
		   ypos++;
		   break;
	case Left: if (animate) {
			putchr(OANS_LEAD, FALSE);
			putchr(OANS_LEFT, FALSE);
			xpos--;
		    }
		   else putchr(BACKSP, TRUE);
		   break;
       case Right: if (animate) {
			putchr(OANS_LEAD, FALSE);
			putchr(OANS_RGHT, FALSE);
			xpos++;
		   }
		   else putchr(SPACE, TRUE);
		   break;
	}
    }
    checkpos();
} 


void translate()                        /*  Main File Translate routine  */
{ 
    unsigned char ch;
    register bool cmdfound;

    putchr(OANS_NMRE, FALSE);
    while (!feof(infile)) {
	if ( (ch=getchr()) == ANSI_LEAD) {          /*  Parse for ANSI lead-in  */
	    if ((ch = getchr()) == ANSI_CMD) {      /*  Found ANSI command  */
		ccbuff = 0;
		codebuff[0] = 0;
		do {
		    codebuff[ccbuff] = (ch = getchr());
		    codebuff[++ccbuff] = 0;
		    cmdfound = TRUE;
		    switch (ch) {
			case 'm':   setattr();          /*  Set graphics attribute  */
				    break;
			case 'H':   setposn();          /*  Set cursor position  */
				    break;
			case 'J':   clrscreen();        /*  Clear the screen  */
				    break;
			case 'K':   clrtoend();         /*  Clear to end of line  */
				    break;
			case 's':   active=FALSE;       /*  Save current cursor position  */
				    break;
			case 'u':   active=TRUE;        /*  Restore saved position  */
				    break;
			case 'A':   cursormove(Up);     /*  Cursor up line  */
				    break;
			case 'B':   cursormove(Down);   /*  Cursor down line  */
				    break;
			case 'D':   cursormove(Left);   /*  Cursor left character  */
				    break;
			case 'C':   cursormove(Right);   /*  Cursor right character  */
				    break;
			 default:   cmdfound = FALSE;
				    break;
		    }
		} while (!(cmdfound));
	    }
	    else {
		putchr(ANSI_LEAD, TRUE);            /*  Ignore non-ANSI ESC  */
		putchr(ch, TRUE);
            } 
        } 
	else putchr(ch, TRUE);
    }
    putchr(OANS_MORE, FALSE);
}


main(argc, argv)
int argc;
char *argv[];
{
    register int j, i;
    bool shift=FALSE;

    copyright();
    for (i=argc; --i > 0; ) {                    /* Animate option */
	if (stricmp(argv[i], "-a") == 0) {
	    fprintf(stderr,"Animation option selected\n");
	    animate = TRUE;
	    shift = TRUE;
	}                                       /* Static option */
	if (stricmp(argv[i], "-s") == 0) {
	    fprintf(stderr,"Static option selected\n");
	    animate = FALSE;
	    shift = TRUE;
	}
	if (shift) {
	    for (j=i; j < argc; j++) argv[j] = argv[j+1];
	    shift = FALSE;
	    argc--;
	}
    }
    fputc('\n',stderr);
    margc = argc;            /* Transfer these to global variables */
    margv = argv;
    if (getfilenames()) {
	translate();
	fclose(infile);
	fclose(outfile);
    }
} 
